@@ -13,6 +13,7 @@ class MetricsLayerConnectionError(Exception):
 class ConnectionType:
     snowflake = Definitions.snowflake
     bigquery = Definitions.bigquery
+    redshift = Definitions.redshift
 
 
 class BaseConnection:
@@ -81,11 +82,59 @@ class SnowflakeConnection(BaseConnection):
         return {key: attributes.get(key) for key in sort_order if attributes.get(key) is not None}
 
 
+class RedshiftConnection(BaseConnection):
+    def __init__(
+        self,
+        name: str,
+        host: str,
+        username: str,
+        password: str,
+        port: int = 5439,
+        database: str = None,
+        schema: str = None,
+        **kwargs,
+    ) -> None:
+        self.type = ConnectionType.redshift
+        self.name = name
+        self.host = host
+        self.port = port
+        self.username = username
+        self.password = password
+        self.database = database
+        self.schema = schema
+
+    def to_dict(self):
+        base = {
+            "name": self.name,
+            "host": self.host,
+            "port": self.port,
+            "username": self.username,
+            "password": self.password,
+            "type": self.type,
+        }
+        if self.database:
+            base["database"] = self.database
+        if self.schema:
+            base["schema"] = self.schema
+        return base
+
+    def printable_attributes(self):
+        attributes = deepcopy(self.to_dict())
+        attributes.pop("password")
+        attributes["name"] = self.name
+        sort_order = ["name", "type", "host", "port", "username", "database", "schema"]
+        return {key: attributes.get(key) for key in sort_order if attributes.get(key) is not None}
+
+
 class BigQueryConnection(BaseConnection):
-    def __init__(self, name: str, credentials: str, **kwargs) -> None:
+    def __init__(
+        self, name: str, schema: str = None, credentials: str = None, keyfile: str = None, **kwargs
+    ) -> None:
         self.type = ConnectionType.bigquery
         self.name = name
-        self.credentials = self._convert_json_if_needed(credentials, kwargs)
+        self.schema = schema
+        creds_to_use = credentials if credentials else keyfile
+        self.credentials = self._convert_json_if_needed(creds_to_use, kwargs)
         self.project_id = self.credentials["project_id"]
 
     def to_dict(self):
